#!/usr/bin/perl

# NDMP Index File Parser
# Takes a backup number and an NDMP index file as arguments and adds the files
# and directories therein to the FILES_TABLE and DIRS_TABLE, respectively. If
# no index file is provided, an attempt will be made to mount the backup,
# extract the index file, and proceed.
#
# The verbose (-v) and debug (-d) options send their output to STDERR. If
# calling this script from elsewhere, consider using the -v option and
# redirecting STDERR to append to <LOG_DIR>/ndmp_index_parser.log.
#
# This script returns 0 on success and the error code returned by the DBD::Pg
# module if an error occurs. If any file names were not able to be placed in
# the database (typically due to character encoding problems), the script
# returns 254. This is not, strictly speaking, a failure, and this exit code
# should be checked for when calling.
#
# NOTE CentOS 5 and CentOS 6 have different versions of the DBD::Pg module. The
# CentOS 5 version is heavily deprecated and requires slightly different usage
# than the CentOS 6 version. The NEW_DBD_VERS constant is used to differentiate
# between the two usages.

use strict;
use warnings;
use DBI;
use DBD::Pg;
use URI::Escape;
use Getopt::Std;
use POSIX;

use constant {
	BPUTIL => "/usr/bp/bin/bputil",
	MASTER_INI => "/usr/bp/bpinit/master.ini",
	DATABASE => "bpdb",
	DB_USER => "postgres",
	DIRS_TABLE => "bp.dirs_non_ctar",
	FILES_TABLE => "bp.backup_files_non_ctar",
	END_BACKUP_NO => 2147483647,    # Convention for non-chained backups
	FT_REG => 0,                    # Internal file type identifier for regular files
	FT_DIR => 53,                   # Internal file type identifier for directories
	FT_CONTAINER => 57,             # Internal file type identifier for containers
	NDMP_MOUNT_PATH => "/backups/ndmp", # Temporary location to mount NDMP backups
	VOLUME => '@@@:',               # Special volume container entry
	NEW_DBD_VERS => 2.0,            # Earliest known version to use new functions
	VERSION => "1.4.4",             # Version of this script
};


my (%opts, $backup_no, $index_file, $mounted);

# Database handler
my $dbh;

# script's log level 0 = Quiet, 1 = Verbose, 2 = Debug
my $LOG_LEVEL = 0;

# Hash of binary scalars with packed info about node/dir/file
my %lookup = ();


# Print usage and exit
sub usage_and_exit()
{
	print STDERR "This script parses NDMP index files and adds the names ",
		"and file attributes of files and directories in the backup to ",
		"a database.\n",
		"Usage: $0 [-h] [-l <LOG_LEVEL>] -b <BACKUP_NUMBER> [-f <FILE>]\n",
		"-h\tPrint this help and exit.\n",
		"-l LOG_LEVEL\t0 = Quiet (errors and warnings only), 1 = Verbose, 2 = Debug.\n",
		"-b NUM\tBackup number.\n",
		"-f FILE\tNDMP index file. If not provided, the file will be inferred ",
		"based on the backup number.\n";
	exit(1);
}

# Return a timestamp for the current time as a string
sub timestamp()
{
	return strftime "%F %T", localtime;
}

# Prints a log message to STDERR including a timestamp. Takes two arguments,
# the log level and the log message. The message will only be printed if the
# $LOG_LEVEL is greater than or equal to the provided log level.
# logging(0, "Message") will always print, but logging(1, "Message") will only
# print if $LOG_LEVEL is set to 1 or higher.
#  usage: same as print but with arg is loglevel: logging(2, "my debug message value: ", value, " value2 ", value2);
sub logging
{
	my $level = shift;
	if (defined($level) && $LOG_LEVEL >= $level && scalar(@_) > 0) {
		print STDERR "[", timestamp(), "] ", @_, "\n";
	}
}

# Connect to the database
sub db_connect()
{
	logging(1, "Connecting to database...");
	$dbh = DBI->connect("dbi:Pg:dbname=" . DATABASE, DB_USER, "")
			or die "Could not connect to database server\n";
}

# Disconnect from the database
sub db_disconnect()
{
	logging(1, "Disconnect from database\n");
	$dbh->disconnect();
	undef $dbh;
}

# Mount backup
sub mount()
{
	logging(1, "Mount backup $backup_no");
	system("/usr/bp/bin/cmc_dump_mount", "--path=" . NDMP_MOUNT_PATH, "mount",
			$backup_no) == 0 or die "Failed to mount backup $backup_no\n";
	$index_file = NDMP_MOUNT_PATH . "/_dump/$backup_no/index";
	$mounted = 1;
}

# Unmount backup if $mounted
sub unmount()
{
	if ($mounted) {
		logging(1, "Unmount backup $backup_no");
		system("/usr/bp/bin/cmc_dump_mount", "--path=" . NDMP_MOUNT_PATH, "destroy",
				$backup_no) == 0 or logging(0, "Failed to unmount backup $backup_no");
	}
}

# Prepare to exit, especially in case of an error. Call before dying.
sub cleanup()
{
	db_disconnect();
	unmount();
}

# Initialize the script (read options, connect to database)
sub init()
{
	my $opt_string = 'hl:f:b:';
	getopts("$opt_string", \%opts) or usage_and_exit();
	usage_and_exit() if $opts{h} or not %opts;
	$index_file = $opts{f};
	$backup_no = $opts{b};
	# Set LOG_LEVEL from command line or master.ini [NDMP] ParserLogLevel, preferring command line
	if (defined($opts{l})) {
		$LOG_LEVEL = $opts{l};
	} else {
		open my $conf, '-|', BPUTIL . " -g NDMP ParserLogLevel $LOG_LEVEL " . MASTER_INI;
		$LOG_LEVEL = <$conf>;
		close $conf;
	}
	# Ensure LOG_LEVEL is at least 0
	if ($LOG_LEVEL < 0) {$LOG_LEVEL = 0};
	$mounted = 0;
	db_connect();
}

# Get the full filepath from a record
sub get_path_from_node($)
{
	my $record = $_[0];
	my $path = "";

	while (defined($record) && $record->{id} != 2) { # The root node is always 2
		$path = '/' . $record->{name} . $path;
		$record = get_record($record->{parent});
	}

	return $path;
}


# return 1 if record exists
sub is_exists($)
{
	# defined() is used here because deleted items has undef value
	return defined($lookup{$_[0]});
}


#### storage interface
# write record; return 1 on success
sub write_record($)
{
	my ($record) = @_;

	# pack to save memory, around 3.33 times less memory but 1.5 times slower when compared against hash of arrays
	my $packed = pack(
		'QlQCQQZ*',
		$record->{parent},
		$record->{mod_date},
		$record->{block_offset},
		$record->{type},
		$record->{dir_no},
		$record->{size},
		$record->{name},
		);

	# save to hash
	$lookup{$record->{id}} = $packed;

	return 1;
}

# decode record; on errors return undef
sub decode_record($$)
{
	my ($id, $packed) = @_;

	my ($parent, $mod_date, $block_offset, $type, $dir_no, $size, $name) = unpack('QlQCQQZ*', $packed);
	if (defined($name)) {
		my $record = {
			id => $id,
			parent => $parent,
			mod_date => $mod_date,
			block_offset => $block_offset,
			type => $type,
			dir_no => $dir_no,
			size => $size,
			name => $name,
			};
		return $record;
	}

	return undef;
}

# update record return 1 on success
sub update_record($)
{
	return write_record($_[0]);
}

# get record by record id; errors or on missing record return undef
sub get_record($)
{
	my ($id) = @_;

	# get value from hash
	my $packed = $lookup{$id};
	if (defined($packed)) {
		# and if set decode it to record
		return decode_record($id, $packed);
	}

	return undef
}

# create root record if no record with same id doesn't exists, return 1 on success
sub create_root_record($)
{
	my ($record) = @_;

	my $exists = is_exists($record->{id});
	if(! $exists) {
		return write_record($record);
	}
	return 1;
}

# create and save new record without and conditions return 1 on success
sub create_node_record($)
{
	return write_record($_[0]);
}

# update existing node record to file record; return 1 on success or node doesn't exists
sub update_file_record($)
{
	my ($record) = @_;

	my $old = get_record($record->{id});
	if ($old) {
		# copy updated values
		$old->{size} = $record->{size};
		$old->{mod_date} = $record->{mod_date};
		$old->{block_offset} = $record->{block_offset};

		# and save
		return update_record($old);
	}
	return 1;
}

# update existing node record to dir record; return 1 on success or node doesn't exists
sub update_dir_record($)
{
	my ($record) = @_;

	my $old = get_record($record->{id});
	if ($old) {
		# copy updated values
		$old->{mod_date} = $record->{mod_date};
		$old->{block_offset} = $record->{block_offset};
		$old->{type} = $record->{type};
		$old->{dir_no} = $record->{dir_no};

		# and save
		return update_record($old);
	}
	return 1;
}

# update existing node record to set dir_no; return 1 on success on errors or record doesn't exist returned undef
sub set_dir_no($$)
{
	my ($record, $dir_no) = @_;

	# copy updated values
	$record->{dir_no} = $dir_no;

	# and save
	return update_record($record);
}

# get dir_no by record id; return dir_no or undef of errors or record doesn't exist
sub get_dir_no($)
{
	my ($id) = @_;

	my $record = get_record($id);
	if ($record) {
		return $record->{dir_no};
	}
	return undef;
}
### End of storage interface


sub mark_record_as_deleted($)
{
	my ($id) = @_;
	# we can't directly remove item from %lookup when loop of while(each()) {}
	# so remove content of item
	$lookup{$id} = undef;
}

# return 0 on success
sub main()
{
	$Getopt::Std::STANDARD_HELP_VERSION = 1;

	init();

	# Unless an index file was specified, and exists, attempt to mount the backup
	# and use the index file contained within
	mount() unless ($opts{f} and -e $opts{f});

	logging(1, "Index File: $index_file, Backup No.: $backup_no");

	# Prepare INSERT/SELECT statements for DIRS_TABLE
	my $sth = $dbh->prepare("INSERT INTO " . DIRS_TABLE . " (dname) SELECT ? " .
		"WHERE NOT EXISTS (SELECT TRUE FROM " . DIRS_TABLE . " WHERE dname = ?)");
	my $sth2 = $dbh->prepare("SELECT dir_no FROM " . DIRS_TABLE .  " WHERE dname = ?");

	# The blank directory entry should be added by the schema definition, but is
	# included here for safety and consistencty.
	$sth->execute('', '');
	$sth2->execute('');
	my $volume_dir_no = $sth2->fetchrow_array();

	logging(2, "\$volume_dir_no is $volume_dir_no");

	# Give cmc_dump_mount time to finish mounting. Using sleep is not ideal, but
	# all the implicit forking and asynchronous commands means waiting for signals
	# can potentially hang indefinitely; this approach allows us to set a timeout.
	my $time = 1800;
	while ($mounted and not -e $index_file and $time) {
		if ($time % 10 == 0) {
			logging(1, "Mounting backup $backup_no, $time seconds remaining");
		}
		$time -= 1;
		sleep 1;
	}

	# FIXME: whats about file encoding ???
	if (open(NDMP_INDEX, "<", $index_file)) {
		logging(1, "Parse $index_file...");

		# Read through the index file and to temporary files
		while (<NDMP_INDEX>) {
			chomp;
			# Type - Parent - Name - Host Type? - Node
			if ($_ =~ /^DHd (\d+) (.+) UNIX (\d+)/) {
				# Node
				if ($2 ne '.' and $2 ne '..') {
					my $node = {
						id => $3,
						parent => $1,
						mod_date => -1,
						block_offset => 0,
						type => FT_REG,
						dir_no => 0,
						size => 0,
						name => uri_unescape($2),
						};
					create_node_record($node);
					undef $node;
				}
			} elsif ($_ =~ /^DHn (\d+) UNIX fd m\d+ u\d+ g\d+ tm(\d+) @(\d+)/) {
				# Directory
				my $dir = {
					id => $1,
					mod_date => $2,
					block_offset => $3,
					type => FT_DIR,
					dir_no => $volume_dir_no,
					};
				update_dir_record($dir);
				undef $dir;

			} elsif ($_ =~ /^DHn (\d+) UNIX f- m\d+ u\d+ g\d+ s(\d+) tm(\d+) @(\d+)/) {
				# File
				my $file = {
					id => $1,
					mod_date => $3,
					block_offset => $4,
					size => $2,
					};
				update_file_record($file);
				undef $file;
			} elsif ($_ =~ /^DHr (\d+)$/) {
				# Root
				my $root = {
					id => $1,
					parent => 0,
					mod_date => 0,
					block_offset => 0,
					type => FT_DIR,
					dir_no => 0,		# would be $volume_dir_no ?
					size => 0,
					name => '/',
					};
				create_root_record($root);
				undef $root;
			} else {
				# No matches
				logging(2, "UNHANDLED record $_");
			}
		}
		logging(1, "Done parsing");
		close(NDMP_INDEX);
	} else {
		cleanup();
		undef $sth;
		undef $sth2;
		die "Cannot open '$index_file': $!\n";
	}

	# Parsing finished, no need to leave the backup mounted
	unmount();

	# Add all directory entries to DIRS_TABLE
	logging(1, "Insert directories...");
	$dbh->begin_work();

	my $dir_counter = 0;
	my ($id, $value, $record, $path);

	while (($id, $value) = each %lookup) {
		next if ! defined($value);

		$record = decode_record($id, $value);
		# Only run on directories
		next if (!defined($record) || $record->{type} != FT_DIR);

		my $path = VOLUME . get_path_from_node($record) . "/";

		# Add directory if unique
		$sth->execute($path, $path);
		if ($sth->err) {
			logging(0, "ERROR: Failed to add directory '$path': " . $dbh->errstr);
		}

		# Get DIR_NO to update the current item in the lookup table
		$sth2->execute($path);
		if ($sth2->err) {
			logging(0, "ERROR: Could not retrieve DIR_NO for '$path': " . $dbh->errstr);
			# This will make the entry unusable, remove it from the lookup table
			mark_record_as_deleted($id);
		} else {
			set_dir_no($record, $sth2->fetchrow_array());
			$dir_counter++;
		}
	}

	$dbh->commit();

	if ($dbh->err) {
		logging(0, "Failed to commit directories to database: " . $dbh->errstr);
		undef $sth2;
		undef $sth;
		db_disconnect();
		return $dbh->err;
	}

	logging(1, "Done inserting directories");
	logging(2, "$dir_counter directories has been added/updated to " . DIRS_TABLE);

	undef $sth2;
	undef $sth;

	logging(2, "number of records in hash ". scalar(keys(%lookup)) .  " dirs $dir_counter");

	# Generate tuples to COPY into FILES_TABLE
	logging(1, "Insert files...");
	$dbh->do("COPY " . FILES_TABLE . "(backup_no, end_backup_no, dir_no, fname,
		type, mod_date, blk_offset, size, blocks) FROM STDIN");

	my $text = "$backup_no\t" .
		END_BACKUP_NO . "\t" .
		"$volume_dir_no\t" .
		VOLUME . "\t" .
		FT_CONTAINER . "\t" .
		"0\t" .
		"0\t" .
		"0\t" .
		"0\n";

	my $old_school_db = $DBD::Pg::VERSION < NEW_DBD_VERS;

	# Add "@@@:" entry manually
	if ($old_school_db) {
		$dbh->pg_putline($text);
	} else {
		$dbh->pg_putcopydata($text);
	}

	# Add all other entries
	my $file_counter = 1;
	my $dir_no;
	while(($id, $value) = each %lookup) {
		next if $id == 0 || !defined($value);

		$record = decode_record($id, $value);
		# Some items may need to be skipped (such as unmodified files from incrementals)
		next if (!defined($record) || $record->{parent} == 0 || $record->{mod_date} == -1);

		$dir_no = get_dir_no($record->{parent});
		next if ! defined ($dir_no);

		$file_counter++;

		$text = "$backup_no\t" .
			END_BACKUP_NO . "\t" .
			"$dir_no\t" .
			"$record->{name}\t" .
			"$record->{type}\t" .
			"$record->{mod_date}\t" .
			"$record->{block_offset}\t" .
			"$record->{size}\t" .
			"0\n";

		if ($old_school_db) {
			$dbh->pg_putline($text);
		} else {
			$dbh->pg_putcopydata($text);
		}
	}

	if ($old_school_db) {
		$dbh->pg_endcopy();
	} else {
		$dbh->pg_putcopyend();
	}


	my $rv = 0;

	# If there was an error exiting COPY mode, the transaction will fail to commit;
	# this results in a backup which has no file names in the database, making
	# selective restores impossible. Attempt to insert the files one at a time
	# using prepared statements to get as many file names as possible into the
	# database and warn the user that some files may be unavailable for selective
	# restore. Note that this is MUCH (~60 times) slower and should be used only as
	# a last resort.
	if ($dbh->err) {
		$file_counter = 0;
		logging(0, "WARNING: Backup $backup_no (or some files in it) may be ".
			"ineligible for selective restore. DB Error: " . ($dbh->errstr || ''));
	
		# The CentOS 5 version of the DBD::Pg module does not reset the internal
		# COPY flag if pg_endcopy() fails. The database connection needs to be
		# reset to issue further commands.
		if ($old_school_db) {
			db_disconnect();
			db_connect();
		}
	
		$sth = $dbh->prepare("INSERT INTO " . FILES_TABLE . " (backup_no,
			end_backup_no, dir_no, fname, type, mod_date, blk_offset, size, blocks)
			VALUES(?, ?, ?, ?, ?, ?, ?, ?, 0)");
	
		# Add "@@@:" entry manually
		$sth->execute($backup_no, END_BACKUP_NO, $volume_dir_no, VOLUME, FT_CONTAINER, 0, 0, 0);
	
		while(($id, $value) = each %lookup) {
			next if ($id == 0 || !defined($value));

			$record = decode_record($id, $value);
			next if (!defined($record) || $record->{parent} == 0 || $record->{mod_date} == -1);

			$dir_no = get_dir_no($record->{parent});
			next if ! defined($dir_no);

			$sth->execute($backup_no, END_BACKUP_NO,
				$dir_no,
				$record->{name},
				$record->{type},
				$record->{mod_date},
				$record->{block_offset},
				$record->{size});
	
			if ($sth->err and $sth->err == 7) {
				# A return value of 7 indicates a fatal error was returned, the
				# last query failed. This is probably due to a non-UTF-8 filename,
				# so let's log the filename.
				my $bad_path = get_path_from_node($record);
				logging(0, "WARNING: Failed to add file '$bad_path': " . $dbh->errstr);
			} else {
				$file_counter++;
			}
		}

		$rv = 254; # Let caller know something went wrong and to flag this backup with a warning.
	}
	undef %lookup;

	logging(1, "Done inserting files");
	logging(2, "$file_counter files/directories have been inserted into " . FILES_TABLE);

	db_disconnect();

	return $rv;
}

my $code = main();
exit($code);
